/* RestXmpl.cmd - trivial BackEmUp Rexx Restoration Example

   This demonstrates how the information is stored in the backup media files
   of the "BackEmUp" and "BackEmUI" programs.

   It does not have all the Bells and Whistles of the full compiled C program,
   nor does it have the intelligence to properly deal with chunked files.
   It will not automatically re-create the original path information.
   Also, as it does not integrate into the control file, it does not know
   how to invoke the user exits for file Encryption nor Compression.  Places
   where those exits would be invoked from are noted within this example.

   This is for educational purposes as an example of to interpret the backup
   media files, and to ease fears that your backup data locked into a data
   format you can't understand. */

/* get the invocation parameters */
parse arg fn as .
if fn = '' | fn = '-?' | fn = '?' then do
   say 'Syntax:'
   say '   1st argument is the backup media file name to open.  It is required.'
   say '   2nd argument is option and signals if you want to recreate the'
   say '           source using the backup media file.'
   say '       if the argument is a wild card *,'
   say '          then the original unqualified name is used for restoration.'
   exit 100
end

/* make sure we have access to the RexxUtility functions */
call RxFuncAdd "SysLoadFuncs", RexxUtil, "SysLoadFuncs"
call SysLoadFuncs

header = ''
status = stream(fn,'c','open read')
if left(status,5) = 'READY' then do
   /* read some of the file to access the imbedded header information */
   rsize = stream(fn,'c','query size')
   if rsize > 512 then rsize = 512   /* plenty big for max file name & info */
   data = charin(fn,1,rsize)
   eHeader = pos(d2c(0),data)     /* header string always ends with a binary 0 */

trace("?R")
   /* check the internal header for what we expect */
   checkChar = substr(data,2,1)
   if checkChar \= ' ' then do
      if substr(data,eHeader+1,1) = '1a'x then do
         /* the header was trivially scrambled.  Note, the actual file contents
            are encrypted with a stronger algorithm via the user exit */
         eHeader = eHeader + 1      /* there 1 more real byte in the header */
         do l = 1 to eHeader
            newChar = bitxor(substr(data,l,1),'1a'x)    /* 1A is used as EOF marker for 'type' command */
            data = overlay(newChar,data,l,1)
         end
         checkChar = substr(data,2,1)
         if checkChar \= ' ' then do
            say 'This does not look like an encrypted backup file to me!'
            exit (8)
         end
      end
      else do
         say 'This does not look like a backup file to me!'
         exit (8)
      end
   end
trace("?R")

   /* position just past the file header */
   checkChar = charin(fn,eHeader,1)

   status = stream(fn,'s')
   if left(status,5) = 'READY' then do
      header = data
      parse var header type adate '"' longname '"' shortname . '(' generation ')'

      if type = 'D' then lType = 'Directory'
      else if type = 'F' then lType = 'File'
      else do
         say 'This does not look like a properly created backup file to me!'
         exit (8)
      end

      say lType '"'||longname||'"'
      if shortname \= '' then say '    also known as "'||shortname||'"'

      /* if option to restore it bassed on the original name */
      if as = '*' then do    /* get just the name portion of original spec */
         if type = 'F'
            then as = filespec('NAME',longname)
            else as = filespec('NAME',left(longname,length(longname)-1))
      end

      parse var generation . ',' bname ',' attr ',' fdate ',' fsize ',' eaSize ',' esSum ',' bdate .
      if bname \= translate(filespec('name',fn)) then do
         say ' *** inconsistent internal name "'||bname||'" found.  Expecting' fn
      end

      /* examine the attributes of the file */
      szAttr = ''
      a1 = substr(attr,1,1)
      if c2x(bitand(a1,'40'x)) = '00' then do   /* this bit MUST be set to be valid! */
         say 'The embedded header indicates this is not a valid backup file!'
         exit (8)
      end
      if c2x(bitand(a1,'01'x)) \= '00' then szAttr = szAttr||', Archive bit set'
      if c2x(bitand(a1,'02'x)) \= '00' then szAttr = szAttr||', Readonly'
      if c2x(bitand(a1,'04'x)) \= '00' then szAttr = szAttr||', Subdirectory'
      if c2x(bitand(a1,'08'x)) \= '00' then szAttr = szAttr||', Hidden'
      if c2x(bitand(a1,'10'x)) \= '00' then szAttr = szAttr||', System'
      say '  Original file attributes: ' substr(szAttr,2)

      /* examine the backup options of the file */
      szOpt = ''
      a2 = substr(attr,2,1)
      if c2x(bitand(a2,'04'x)) \= '00' then szOpt = szOpt||', Encrypted'
      if c2x(bitand(a2,'02'x)) \= '00' then szOpt = szOpt||', Compressed'
      if szOpt \= '' then say '  Backup media file is' substr(szOpt,2)

      if eaSize \= 0 then do
         say '  Extended Attributes are present, ('||eaSize 'bytes)'
      end
      if type = 'F' then say '  File size:' fsize 'data bytes'

      /* if wanting to restore it */
      if as \= '' then do

         /* if it was encrypted */
         if c2x(bitand(a2,'04'x)) \= '00'  then do
            /* write out the encrypted information without the header */
            dsize = stream(fn,'c','query size') - eHeader
            encryptedData = charin(fn,eHeader+1,dsize)
            rc = charout(as,encryptedData,1)

            /* here's what other steps are required */
            say 'First you must invoke your decyption routine to decrypt' as
            if c2x(bitand(a2,'02'x)) \= '00' then say 'Then invoke your decompression routine on that result.'
               else if eaSize > 0 then do
               say 'Then separate the contents from the extended attributes:'
               say '   The first 4 bytes indicate the length of the file contents portion.'
               say '   The final' eaSize 'bytes are the Extended Attributes.'
            end

            exit (4)  /* that's all we can do here */
         end

         /* if it was compressed */
         if c2x(bitand(a2,'02'x)) \= '00'  then do
            /* write out the compressed information without the header */
            csize = stream(fn,'c','query size') - eHeader
            compressedData = charin(fn,eHeader+1,csize)
            rc = charout(as,compressedData,1)

            say 'Now you must invoke your decompression routine on' as

            exit (4)  /* that's all we can do here */
         end

         /* access the EAs and contents */
         if eaSize \= 0 then do
            /* lets check the consistency markers within the file.
               the next 4 bytes after the header should be the number of
               databytes in the file.  (a 'chunk' is always less than 2 gig) */
            checkFileSizeString = charin(fn,eHeader + 1, 4)
            checkFileSize = 0
            do l = length(checkFileSizeString) to 1 by -1
               checkFileSize = checkFileSize * 256 + c2d(substr(checkFileSizeString,l,1))
            end
            eHeader = eHeader + 4     /* skip past the 4 byte length */
            if checkFileSize <> fsize then do
               say 'Likely error in internal file stucture for file data.'
               exit (16)
            end

            /* and as there are EAs present, check for their length consistency */
            EA_Offset = eHeader + 1 + checkFileSize
            checkEASizeString = charin(fn,EA_Offset, 4)
            checkEASize = 0
            do l = length(checkEASizeString) to 1 by -1
               checkEASize = checkEASize * 256 + c2d(substr(checkEASizeString,l,1))
            end
            if checkEASize <> eaSize then do
               say 'Likely error in internal file stucture for EA data.'
               exit (16)
            end

            /* read the Extended Attributes from toward the end of the file */
            EAs = charin(fn,eHeader + 1 + fsize + 4 ,eaSize)
         end

         /* read the data portion of the file */
         fdata = charin(fn,eHeader + 1,fsize)

         if type = 'D' then do
            /* create the subdirectory */
            'mkdir "'||as||'"'
         end
         else do
            /* create the data portion of the restored file */
            rc = charout(as,fdata,1)
            rc = stream(as,'c','close')
         end

         /* if Extended Attributes were present, join them to the file */
         if eaSize \= 0 then do
            /* we need a temporary file to store the EAs for joining */
            EAPart = SysTempFileName('$TMP???.@??')
            rc = charout(EApart,EAs,1)
            rc = stream(EApart,'c','close')

            /* join them and NOT keep the temporary file */
            '@eautil' as EApart '/J /O'
            if rc \= 0
               then say ' ** Error rejoining the extended attributes from' EAPart
         end

         say 'restored as' as
      end
   end
End
else do
   say 'Unexpected error opening' fn
   exit 12
end
status = stream(fn,'c','close')
